from z3 import *
from symbolic_types import SymbolicType, get_symbolic_expr


class OutOfBuget(Exception):
    pass


class UndefinedOutcomeException(Exception):
    pass


class SymbolicTree:
    def __init__(self, expr):
        self.expr = expr
        self.inv = True
        self.parent = None
        self.true_child = None
        self.false_child = None
        self.output = None

    def set_output(self, result: SymbolicType):
        self.output = result

    def set_true_child(self, child):
        self.true_child = child
        child.parent = self

    def set_false_child(self, child):
        self.false_child = child
        child.parent = self

    def get_true_child(self):
        return self.true_child

    def get_false_child(self):
        return self.false_child

    def find_cumulative_expression(self):
        return And(self.find_cumulative_expression_list())

    def find_cumulative_expression_list(self):
        if self.parent is None:
            return [And(self.expr, self.inv)]
        else:
            results = self.parent.find_cumulative_expression_list()
            results.append(And(self.expr, self.inv))
            return results

    def size(self):
        if self.true_child is None and self.false_child is None:
            return 1
        else:
            t_size = 0 if self.true_child is None else self.true_child.size()
            f_size = 0 if self.false_child is None else self.false_child.size()
            return t_size + f_size

    def is_leaf(self):
        return self.true_child is None and self.false_child is None

    def is_terminal(self):
        return self.is_leaf() and self.output is not OutOfBuget and \
            not isinstance(self.output, Exception)

    def get_symbolic_summary(self, output_var):
        """
        :param output_var: the symbolic variable for representing the output value
        :return: a symbolic summary that defines the input / output relation for the explored function.
        """
        if self.is_leaf():
            if self.is_terminal():
                return Implies(self.expr, And(self.inv, output_var == get_symbolic_expr(self.output)))
            elif isinstance(self.output, Exception):
                # The function is not defined for the path
                return Implies(self.expr, False)
            elif self.output is OutOfBuget:
                # Note: exploration is incomplete on the current path, we do not yet have enough information for
                # summary,
                return Implies(self.expr, self.inv)
            else:
                raise UndefinedOutcomeException()
        else:
            summary_t = self.true_child.get_symbolic_summary(output_var) if self.true_child is not None else True
            summary_f = self.false_child.get_symbolic_summary(output_var) if self.false_child is not None else True
            return Implies(self.expr, And(self.inv, summary_t, summary_f))
