import inspect

from z3 import *

import symbolic_types
from execution_tree import SymbolicTree, OutOfBuget, UndefinedOutcomeException
from symbolic_types import SymbolicBool, SymbolicInteger, SymbolicList


class SymbolicPath():
    def __init__(self, solver: Solver):
        # root of the exploration
        self.root = SymbolicTree(True)

        self.solver = solver

        # current exploration frontier
        self.current_node = self.root

        # number of branching choice before giving up
        self.depth_limit = 1000
        self.cur_depth = 0

    def add_inv(self, expr):
        self.current_node.inv = And(self.current_node.inv, expr)

    def branch_pc(self, expr, direction):
        if direction:
            if self.current_node.get_true_child() is None:
                true_child = SymbolicTree(expr)
                self.current_node.set_true_child(true_child)
            self.current_node = self.current_node.get_true_child()
        else:
            if self.current_node.get_false_child() is None:
                false_child = SymbolicTree(expr)
                self.current_node.set_false_child(false_child)
            self.current_node = self.current_node.get_false_child()
        self.cur_depth += 1

        self.solver.push()
        self.solver.append(expr)
        if self.cur_depth > self.depth_limit:
            raise OutOfBuget()

    def reset_path(self):
        self.cur_depth = 0
        self.current_node = self.root

    def find_path_alternatives(self):
        """
        :return: a set of path conditions representing the "potential" alternative paths that are
         scheduled to be explored later
        """
        alternative = []
        cur_exprs = self.current_node.find_cumulative_expression_list()
        while self.current_node != self.root:
            parent = self.current_node.parent
            last_choice = cur_exprs.pop()
            if (parent.get_true_child() == self.current_node and parent.get_false_child() is None) or \
                    (parent.get_false_child() == self.current_node and parent.get_true_child() is None):
                cur_exprs.append(simplify(Not(last_choice)))
                alternative.append(And(cur_exprs))
                cur_exprs.pop()
            self.current_node = parent
            self.solver.pop()
        return alternative

    def find_cur_expr(self):
        return self.current_node.find_cumulative_expression()

    def set_result(self, result):
        self.current_node.set_output(result)

    def reset_exploration(self):
        self.root = SymbolicTree(True)
        self.solver = Solver()
        self.current_node = self.root
        # number of branching choice before giving up
        self.depth_limit = 1000
        self.cur_depth = 0


def launch(function, arguments, path):
    try:
        result = function(*arguments)
    except OutOfBuget:
        result = OutOfBuget
    except:
        result = UndefinedOutcomeException()

    return result


class Engine():
    def __init__(self):
        self.cur_path = SymbolicPath(Solver())
        self.open = []
        symbolic_types.Cur_engine = self

    def get_cur_path(self):
        return self.cur_path

    def select_path_seed(self):
        """
        We pick an unexplored alternative path for exploration. Currently, we always
        pick the last alternative. Feel free to modify the path selection heuristics
        """
        if not self.open:
            return False, None
        else:
            return self.open.pop()

    def find_path_alternatives(self):
        """
        feasible alternatives to explore next. Each alternative should be represented
        as a tuple (path_condition, solution) where solution is an satisfying assignment
        to the path_condition. Add the alternatives to the list self.open.
        """
        # We have completed exploring an path, we would like to explore alternatives
        alternative = self.cur_path.find_path_alternatives()
        for alt in alternative:
            if self.cur_path.solver.check(alt) == sat:
                model = self.cur_path.solver.model()
                self.open.append((alt, model))

    def explore(self, func, output_var, print_summary=False):
        arg_names = inspect.signature(func).parameters
        args = [symbolic_input(name) for name in arg_names.values()]
        find_violation = False
        while not find_violation:
            result = launch(func, args, self.cur_path)
            self.get_cur_path().set_result(result)

            if print_summary:
                self.print_path_summary(args, result)

            self.find_path_alternatives()
            path, seed = self.select_path_seed()
            if seed is not None:
                # now modify argument value by seed
                for arg in args:
                    arg.set_value(seed)
                    # value = seed.evaluate(arg.expr)
                    # if isinstance(value, IntNumRef):
                    #     # print(int(str(value)))
                    #     arg.value = int(str(value))
                self.cur_path.reset_path()
                # launch(func, args, self.cur_path)
            else:
                print("Done, we have exhaust all ({}) paths".format(self.cur_path.root.size()))
                summary = self.cur_path.root.get_symbolic_summary(output_var)
                self.cur_path.reset_exploration()
                return summary, [arg.expr for arg in args], output_var

    def print_path_summary(self, args, output):
        print("current path exploration result")
        print("PC: {}".format(And(self.cur_path.find_cur_expr())))
        print("On Input: {}".format(' ,'.join(symbolic_types.full_str(var) for var in args)))
        print("Output: {}".format(symbolic_types.full_str(output)
                                  ))


Cur_engine = Engine()


def get_current_path():
    return Cur_engine.get_cur_path()


def symbolic_input(arg, seed=0):
    if arg.annotation == int:
        return SymbolicInteger(Int(arg.name), seed)
    elif arg.annotation == bool:
        return SymbolicBool(Bool(arg.name), False)
    elif arg.annotation == list:
        return SymbolicList(arg.name, [])


def symbolic_boolean(name, seed=True):
    return SymbolicBool(Bool(name), seed)


def symbolic_assert(expr):
    path = get_current_path()
    if expr:
        path.verdict = True
    else:
        path.verdict = False


def equivalence_checking(func1, func2, output_type):
    """
    returns true if and only if func1 and func2 are functional equivalent
    :param func1:
    :param func2:
    :param output_type: the output type of both func1 and func2
    :return:
    """
    # TODO: Your implementation goes here
    pass


def verify_output_true(function):
    summary, f_in, f_out = compute_functional_summary(function, bool, "output")
    s = Solver()
    s.append(summary)
    # we can verify that the output of my_max is always an upper-bound of the inputs
    s.append(Not(f_out))
    result = s.check()
    if result == sat:
        print(s.model())
    assert s.check() == unsat


def compute_functional_summary(func, output_type, output_name, print_summary=False):
    if output_type is bool:
        out = Bool(output_name)
    elif output_type is int:
        out = Int(output_name)
    else:
        print("Unsupported output")
        return False

    return Cur_engine.explore(func, out, print_summary=print_summary)
