from z3 import *

Cur_engine = None


def set_engine(engine):
    global Cur_engine
    Cur_engine = engine


def get_cur_path():
    assert Cur_engine
    return Cur_engine.get_cur_path()


def get_symbolic_expr(content):
    if isinstance(content, SymbolicInteger):
        return content.expr
    if isinstance(content, SymbolicBool):
        return content.expr
    else:
        return content


def get_symbolic_value(content):
    if isinstance(content, SymbolicInteger):
        return content.value
    if isinstance(content, SymbolicBool):
        return content.value
    else:
        return content


class SymbolicType():
    def __init__(self, expr, value):
        self.value = value
        self.expr = expr

    def set_value(self, model):
        value = model.evaluate(self.expr)
        if isinstance(value, IntNumRef):
            value = int(str(value))
        elif value == self.expr:
            value = 0
        self.value = value


class SymbolicInteger(SymbolicType):

    def __add__(self, other):
        return SymbolicInteger(get_symbolic_expr(self) + get_symbolic_expr(other),
                               get_symbolic_value(self) + get_symbolic_value(other))

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        return SymbolicInteger(get_symbolic_expr(self) * get_symbolic_expr(other),
                               get_symbolic_value(self) * get_symbolic_value(other))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __sub__(self, other):
        return SymbolicInteger(get_symbolic_expr(self) - get_symbolic_expr(other),
                               get_symbolic_value(self) - get_symbolic_value(other))

    def __rsub__(self, other):
        return SymbolicInteger(get_symbolic_expr(other) - get_symbolic_expr(self),
                               get_symbolic_value(other) - get_symbolic_value(self))

    def __divmod__(self, other):
        return SymbolicInteger(get_symbolic_expr(self) / get_symbolic_expr(other),
                               get_symbolic_value(self) / get_symbolic_value(other))

    def __rdivmod__(self, other):
        return SymbolicInteger(get_symbolic_expr(other) / get_symbolic_expr(self),
                               get_symbolic_value(other) / get_symbolic_value(self))

    def __gt__(self, other):
        return SymbolicBool(get_symbolic_expr(self) > get_symbolic_expr(other),
                            get_symbolic_value(self) > get_symbolic_value(other))

    def __lt__(self, other):
        return SymbolicBool(get_symbolic_expr(self) < get_symbolic_expr(other),
                            get_symbolic_value(self) < get_symbolic_value(other))

    def __ge__(self, other):
        return SymbolicBool(get_symbolic_expr(self) >= get_symbolic_expr(other),
                            get_symbolic_value(self) >= get_symbolic_value(other))

    def __le__(self, other):
        return SymbolicBool(get_symbolic_expr(self) <= get_symbolic_expr(other),
                            get_symbolic_value(self) <= get_symbolic_value(other))

    def __eq__(self, other):
        return SymbolicBool(get_symbolic_expr(self) == get_symbolic_expr(other),
                            get_symbolic_value(self) == get_symbolic_value(other))

    def __ne__(self, other):
        return SymbolicBool(get_symbolic_expr(self) != get_symbolic_expr(other),
                            get_symbolic_value(self) != get_symbolic_value(other))

    def __str__(self):
        return str(self.expr)

    def full_str(self):
        return "{} : {}".format(str(self.expr), str(self.value))


class SymbolicBool(SymbolicType):
    def __neg__(self):
        return SymbolicBool(Not(self.expr), not self.value)

    def __and__(self, other):
        return SymbolicBool(And(self.expr, get_symbolic_expr(other)),
                            get_symbolic_value(self) and get_symbolic_value(other))

    def __rand__(self, other):
        return self.__and__(other)

    def __or__(self, other):
        return SymbolicBool(Or(self.expr, get_symbolic_expr(other)),
                            get_symbolic_value(self) or get_symbolic_value(other))

    def __ror__(self, other):
        return self.__or__(other)

    def __bool__(self):
        direction = self.value
        path = get_cur_path()
        if direction:
            path.branch_pc(self.expr, direction)
        else:
            path.branch_pc(Not(self.expr), direction)
        return direction

    def __str__(self):
        return str(self.expr)

    def full_str(self):
        return "{} : {}".format(str(self.expr), str(self.value))


class SymbolicList(SymbolicType):
    # We assume an upper-bound on the list length
    List_MAX_Length = 5

    def __init__(self, expr, value):
        super().__init__(expr, value)
        self.symbolic_len = SymbolicInteger(Int("{}_width".format(str(expr))), len(value))
        # add an invariant restricting the size of the list
        get_cur_path().add_inv((self.symbolic_len <= SymbolicList.List_MAX_Length).expr)
        get_cur_path().add_inv((self.symbolic_len >=  0).expr)
        self.symbolic_content = [SymbolicInteger(Int("{}_item_{}".format(str(expr), i)), 0)
                                 for i in range(0, self.symbolic_len.value)]
        self.current_index = 0

    def __iter__(self):
        """
        TODO:
        Implement the list's iterator
        """
        pass

    def __next__(self):
        """
        TODO:
        Implement the next method to fetch the next element in the list
        """
        pass

    def __len__(self):
        """
        TODO:
        return the size of the list
        """
        pass

    def __getitem__(self, index):
        """
        TODO:
        returns the element of the list at index.
        The input index can be an integer or symbolic integer
        raise indexError if index out of bound
        """
        pass

    def __setitem__(self, index, value):
        """
        TODO:
        set the value for list at given the index.
        The input index can be an integer or symbolic integer
        raise indexError if index out of bound
        """
        pass

    def __eq__(self, other):
        """
        TODO:
        """
        pass

    def set_value(self, model):
        """
        Given a model, populate the symbolic list with concrete value.
        """
        self.symbolic_len.set_value(model)
        self.symbolic_content = [SymbolicInteger(Int("{}_item_{}".format(str(self.expr), i)), 0)
                                 for i in range(0, self.symbolic_len.value)]
        for item in self.symbolic_content:
            item.set_value(model)
        self.value = []
        for i in range(self.symbolic_len.value):
            self.value.append(self.symbolic_content[i].value)


def full_str(target):
    if isinstance(target, SymbolicBool) or isinstance(target, SymbolicInteger):
        return target.full_str()
    else:
        return str(target)
